import unittest
import os
import tempfile
import datadict
import shutil
import ruamel.yaml
from datadict import datadict_helpers
from datadict import datadict_yaml


class TestDataDict(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory to store the test files
        self.temp_dir = tempfile.mkdtemp()

        # Create a test dictionary YAML file
        self.dictionary_file = os.path.join(self.temp_dir, "test_dictionary.yml")
        with open(self.dictionary_file, "w") as file:
            file.write("dictionary:\n")

        # Create an instance of the datadict class with the test dictionary
        self.datadict_instance = datadict.datadict(self.dictionary_file)

    def tearDown(self):
        # Remove the temporary directory and its contents after the test
        shutil.rmtree(self.temp_dir)

    def test_load_dictionary(self):
        # Test loading an existing dictionary
        loaded_dict = self.datadict_instance._load_dictionary()
        self.assertIsInstance(loaded_dict, dict)

        # Test loading a non-existent dictionary
        non_existent_dict_file = os.path.join(self.temp_dir, "non_existent_dict.yml")
        with self.assertRaises(FileNotFoundError):
            self.datadict_instance.dictionary_path = non_existent_dict_file
            self.datadict_instance._load_dictionary()

    def test_create_dictionary(self):
        # Test creating a new dictionary
        new_dict_file = os.path.join(self.temp_dir, "new_dict.yml")
        self.datadict_instance.dictionary_path = new_dict_file
        created_dict = self.datadict_instance._create_dictinary()
        self.assertIsInstance(created_dict, dict)
        self.assertTrue(os.path.exists(new_dict_file))

    def test_format_dictionary(self):
        # Test formatting an existing dictionary
        original_yaml = {
            "dictionary": [
                {"name": "field_1"},
                {"name": "field_2", "description": "field_2_description"},
                {
                    "name": "field_3",
                    "description": "field_3_description",
                    "aliases": ["field_3_alias1"],
                },
            ]
        }
        expected_yaml = {
            "dictionary": [
                {"name": "field_1", "description": "", "aliases": []},
                {
                    "name": "field_2",
                    "description": "field_2_description",
                    "aliases": [],
                },
                {
                    "name": "field_3",
                    "description": "field_3_description",
                    "aliases": ["field_3_alias1"],
                },
            ]
        }
        formatted_yaml = self.datadict_instance._format_dictionary(original_yaml)
        self.assertEqual(expected_yaml, formatted_yaml)

    def test_parse_aliases(self):
        # Test parsing dictionary with no aliases
        test_dictionary = {"dictionary": [{"name": "field1"}, {"name": "field2"}]}
        result = self.datadict_instance._parse_aliases(test_dictionary)
        self.assertEqual(result, ["field1", "field2"])

        # Test parsing dictionary with aliases
        test_dictionary = {
            "dictionary": [
                {"name": "field1", "aliases": ["f1", "alias1"]},
                {"name": "field2"},
            ]
        }
        result = self.datadict_instance._parse_aliases(test_dictionary)
        self.assertEqual(result, ["field1", "f1", "alias1", "field2"])

    def test_insert_dict_item(self):
        test_dict = {"key1": "value1", "key3": "value3"}
        result_dict = self.datadict_instance._insert_dict_item(
            test_dict, "key2", "value2", 1
        )
        expected_dict = {"key1": "value1", "key2": "value2", "key3": "value3"}
        self.assertEqual(result_dict, expected_dict)

    def test_update_existing_field(self):
        # Test updating existing field with description
        model_column = {"name": "field1", "description": "desc1"}
        model = {"name": "model1"}
        file_path = "path/to/file1.yml"
        self.datadict_instance._update_existing_field(model_column, model, file_path)
        self.assertEqual(
            self.datadict_instance.existing_fields,
            [
                {
                    "name": "field1",
                    "description": "desc1",
                    "model": "model1",
                    "file": "path/to/file1.yml",
                }
            ],
        )

        # Test updating existing field without description
        model_column = {"name": "field2"}
        model = {"name": "model2"}
        file_path = "path/to/file2.yml"
        self.datadict_instance._update_existing_field(model_column, model, file_path)
        self.assertEqual(
            self.datadict_instance.existing_fields,
            [
                {
                    "name": "field1",
                    "description": "desc1",
                    "model": "model1",
                    "file": "path/to/file1.yml",
                },
                {"name": "field2", "model": "model2", "file": "path/to/file2.yml"},
            ],
        )

    def test_iterate_dictionary_update(self):
        # Test iterating through model YAML and updating descriptions
        model_yaml = {
            "models": [
                {
                    "name": "model1",
                    "columns": [{"name": "field1", "description": "old_desc"}],
                }
            ]
        }
        self.datadict_instance.dictionary_yml = {
            "dictionary": [{"name": "field1", "description": "new_desc"}]
        }
        updates = self.datadict_instance._iterate_dictionary_update(
            model_yaml, "path/to/model.yml"
        )
        self.assertTrue(updates["updated"])
        self.assertEqual(
            model_yaml["models"][0]["columns"][0]["description"], "new_desc"
        )

        # Test iterating through model YAML without updates
        model_yaml = {
            "models": [
                {
                    "name": "model2",
                    "columns": [{"name": "field2", "description": "desc2"}],
                }
            ]
        }
        self.datadict_instance.dictionary_yml = {
            "dictionary": [{"name": "field1", "description": "new_desc"}]
        }
        updates = self.datadict_instance._iterate_dictionary_update(
            model_yaml, "path/to/model.yml"
        )
        self.assertFalse(updates["updated"])
        self.assertEqual(model_yaml["models"][0]["columns"][0]["description"], "desc2")

    def test_collate_metadata(self):
        # Test collating metadata from existing fields
        existing_fields = [
            {"name": "field1", "model": "model1", "description": "desc1"},
            {"name": "field1", "model": "model2", "description": "desc2"},
            {"name": "field2", "model": "model1", "description": "desc3"},
            {"name": "field2", "model": "model2"},
        ]
        result = self.datadict_instance._collate_metadata(existing_fields)
        expected_result = [
            {
                "name": "field1",
                "description": "",
                "description_versions": ["desc1", "desc2"],
                "models": ["model1", "model2"],
            },
            {"name": "field2", "description": "desc3", "models": ["model1", "model2"]},
        ]
        self.assertEqual(result, expected_result)

    def test_apply_data_dictionary_to_file(self):
        # Test applying data dictionary to valid model file
        self.datadict_instance.dictionary_yml = {
            "dictionary": [{"name": "field1", "description": "new_desc"}]
        }
        model_yaml = {
            "models": [
                {
                    "name": "test_model",
                    "columns": [{"name": "field1", "description": "old_desc"}],
                }
            ]
        }
        model_yaml_file = os.path.join(self.temp_dir, "model_file.yml")
        with open(model_yaml_file, "w") as file:
            self.datadict_instance.yaml.dump(model_yaml, file)
        self.datadict_instance.apply_data_dictionary_to_file(model_yaml_file)
        with open(model_yaml_file, "r") as file:
            updated_yaml = self.datadict_instance.yaml.load(file)
        expected_yaml = {
            "models": [
                {
                    "name": "test_model",
                    "columns": [{"name": "field1", "description": "new_desc"}],
                }
            ]
        }
        self.assertEqual(updated_yaml, expected_yaml)

    def test_apply_data_dictionary_to_path(self):
        # Test applying data dictionary to files in a directory
        self.datadict_instance.dictionary_yml = {
            "dictionary": [{"name": "field1", "description": "new_desc"}]
        }
        model_yaml_file1 = os.path.join(self.temp_dir, "model_file1.yml")
        model_yaml_1 = {
            "models": [
                {
                    "name": "test_model1",
                    "columns": [{"name": "field1", "description": "old_desc"}],
                }
            ]
        }
        model_yaml_file2 = os.path.join(self.temp_dir, "model_file2.yml")
        model_yaml_2 = {
            "models": [
                {
                    "name": "test_model2",
                    "columns": [{"name": "field1", "description": "old_desc"}],
                }
            ]
        }
        with open(model_yaml_file1, "w") as file:
            self.datadict_instance.yaml.dump(model_yaml_1, file)
        with open(model_yaml_file2, "w") as file:
            self.datadict_instance.yaml.dump(model_yaml_2, file)
        self.datadict_instance.apply_data_dictionary_to_path(self.temp_dir)
        with open(model_yaml_file1, "r") as file:
            updated_yaml1 = self.datadict_instance.yaml.load(file)
        with open(model_yaml_file2, "r") as file:
            updated_yaml2 = self.datadict_instance.yaml.load(file)
        expected_yaml1 = {
            "models": [
                {
                    "name": "test_model1",
                    "columns": [{"name": "field1", "description": "new_desc"}],
                }
            ]
        }
        expected_yaml2 = {
            "models": [
                {
                    "name": "test_model2",
                    "columns": [{"name": "field1", "description": "new_desc"}],
                }
            ]
        }
        self.assertEqual(updated_yaml1, expected_yaml1)
        self.assertEqual(updated_yaml2, expected_yaml2)

    def test_collate_output_dictionary(self):
        # Test loading missing fields with existing fields
        self.datadict_instance.existing_fields = [
            {"name": "field3", "model": "model1", "description": "desc1"},
            {"name": "field3", "model": "model2", "description": "desc2"},
            {"name": "field4", "model": "model1", "description": "desc3"},
            {"name": "field4", "model": "model2"},
        ]

        self.datadict_instance.collate_output_dictionary()

        with open(self.datadict_instance.dictionary_path, "r") as file:
            new_dict = self.datadict_instance.yaml.load(file)
        expected_missing_fields = [
            {
                "name": "field3",
                "description": "",
                "description_versions": ["desc1", "desc2"],
                "models": ["model1", "model2"],
            },
            {"name": "field4", "description": "desc3", "models": ["model1", "model2"]},
        ]

        self.assertEqual(new_dict["dictionary"], expected_missing_fields)


class TestHelpers(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory to store the test files
        self.temp_dir = tempfile.mkdtemp()
        self.yaml_obj = ruamel.yaml.YAML()

    def tearDown(self):
        # Remove the temporary directory and its contents after the test
        shutil.rmtree(self.temp_dir)

    def test_open_model_yml_file(self):
        # Test opening a valid model YAML file
        model_yaml_file = os.path.join(self.temp_dir, "model_file.yml")
        model_yaml = {
            "models": [
                {
                    "name": "test_model",
                    "columns": [{"name": "field1", "description": "old_desc"}],
                }
            ]
        }
        with open(model_yaml_file, "w") as file:
            self.yaml_obj.dump(model_yaml, file)
        result = datadict_helpers.open_model_yml_file(self.yaml_obj, model_yaml_file)
        self.assertTrue(result["status"] == "valid")
        self.assertIsInstance(result["yaml"], dict)

        # Test opening an invalid model YAML file
        invalid_model_yaml_file = os.path.join(self.temp_dir, "invalid_model_file.yml")
        with open(invalid_model_yaml_file, "w") as file:
            file.write("invalid_data:\n")
        result = datadict_helpers.open_model_yml_file(
            self.yaml_obj, invalid_model_yaml_file
        )
        self.assertTrue(result["status"] == "invalid")
        self.assertIsNone(result["yaml"])

    def test_check_valid_model_file(self):
        # Test checking a valid model YAML data
        valid_model_yaml = {
            "models": [
                {
                    "name": "model1",
                    "columns": [{"name": "column1", "description": "desc1"}],
                }
            ]
        }
        result = datadict_helpers.check_valid_model_file(valid_model_yaml)
        self.assertTrue(result)

        # Test checking an invalid model YAML data
        invalid_model_yaml = {"invalid_data": "data"}
        result = datadict_helpers.check_valid_model_file(invalid_model_yaml)
        self.assertFalse(result)

    def test_output_model_file(self):
        # Prepare test data
        model_yaml = {
            "models": [
                {
                    "name": "test_model",
                    "description": "Test Model",
                    "columns": [
                        {"name": "col1", "description": "Column 1"},
                        {"name": "col2", "description": "Column 2"},
                    ],
                }
            ]
        }
        test_file_path = "test_output_model.yaml"
        try:
            datadict_helpers.output_model_file(
                self.yaml_obj, test_file_path, model_yaml, False
            )
            self.assertTrue(os.path.exists(test_file_path))
            with open(test_file_path, "r") as f:
                loaded_yaml = self.yaml_obj.load(f)
            self.assertEqual(loaded_yaml, model_yaml)

        finally:
            if os.path.exists(test_file_path):
                os.remove(test_file_path)

    def test_list_directory_files(self):
        extensions = [".yml", ".yaml"]
        yaml_files = [
            "file1.yaml",
            "file2.yml",
            "subdir/file3.yaml",
            "subdir/file4.yml",
            "subdir/subsubdir/file5.yaml",
            "subdir/subsubdir/file6.yml",
            "non_yaml_file.txt",
        ]
        for file in yaml_files:
            file_path = os.path.join(self.temp_dir, file)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, "w") as f:
                f.write("Sample YAML content")
        # Call the function to get the list of YAML files in the temporary directory
        yaml_files_list = datadict_helpers.list_directory_files(
            self.temp_dir, extensions
        )

        # Assert that the function returns the correct list of YAML files
        expected_yaml_files = [
            os.path.join(self.temp_dir, "file1.yaml"),
            os.path.join(self.temp_dir, "file2.yml"),
            os.path.join(self.temp_dir, "subdir/file3.yaml"),
            os.path.join(self.temp_dir, "subdir/file4.yml"),
            os.path.join(self.temp_dir, "subdir/subsubdir/file5.yaml"),
            os.path.join(self.temp_dir, "subdir/subsubdir/file6.yml"),
        ]
        self.assertCountEqual(yaml_files_list, expected_yaml_files)

    def test_sort_model_file(self):
        # Input dictionary with unsorted models and columns
        input_dict = {
            "version": 2,
            "models": [
                {
                    "name": "dmn_clients",
                    "columns": [
                        {"name": "updated_at", "description": ""},
                        {
                            "name": "client_uuid",
                            "description": "Unique identifier for the client",
                            "tests": ["unique", "not_null"],
                        },
                        {
                            "name": "name",
                            "description": "Name of the client",
                            "tests": ["not_null"],
                        },
                        {
                            "name": "code",
                            "description": "Code used to uniquely identify the client",
                            "tests": ["not_null"],
                        },
                        {
                            "name": "board_name",
                            "description": "Name of the board in monday.com",
                        },
                        {
                            "name": "status",
                            "description": "Status of the client",
                            "tests": ["not_null"],
                        },
                    ],
                },
                {
                    "name": "dmn_stories",
                    "columns": [
                        {"name": "valid_from", "description": ""},
                        {"name": "story_id", "description": ""},
                        {"name": "name", "description": ""},
                        {"name": "client_uuid", "description": ""},
                        {"name": "client_name", "description": ""},
                    ],
                },
            ],
        }

        # Expected output dictionary with sorted models and columns
        expected_output = {
            "version": 2,
            "models": [
                {
                    "name": "dmn_clients",
                    "columns": [
                        {
                            "name": "board_name",
                            "description": "Name of the board in monday.com",
                        },
                        {
                            "name": "client_uuid",
                            "description": "Unique identifier for the client",
                            "tests": ["unique", "not_null"],
                        },
                        {
                            "name": "code",
                            "description": "Code used to uniquely identify the client",
                            "tests": ["not_null"],
                        },
                        {
                            "name": "name",
                            "description": "Name of the client",
                            "tests": ["not_null"],
                        },
                        {
                            "name": "status",
                            "description": "Status of the client",
                            "tests": ["not_null"],
                        },
                        {"name": "updated_at", "description": ""},
                    ],
                },
                {
                    "name": "dmn_stories",
                    "columns": [
                        {"name": "client_name", "description": ""},
                        {"name": "client_uuid", "description": ""},
                        {"name": "name", "description": ""},
                        {"name": "story_id", "description": ""},
                        {"name": "valid_from", "description": ""},
                    ],
                },
            ],
        }

        # Call the function to sort the model file
        sorted_output = datadict_helpers.sort_model_file(input_dict)

        # Compare the actual output with the expected output
        self.assertEqual(sorted_output, expected_output)


class TestYaml(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory to store the test files
        self.temp_dir = tempfile.mkdtemp()
        self.yaml_obj = ruamel.yaml.YAML()

    def tearDown(self):
        # Remove the temporary directory and its contents after the test
        shutil.rmtree(self.temp_dir)

    def test_combine_column_lists_no_missing_columns(self):
        # Test when there are no missing columns
        current_yml = {
            "name": "Model1",
            "columns": [
                {"name": "Column1", "type": "int"},
                {"name": "Column2", "type": "str"},
            ],
        }
        expected_yml = {
            "columns": [
                {"name": "Column1", "type": "int"},
                {"name": "Column2", "type": "str"},
            ]
        }
        result = datadict_yaml.combine_column_lists(current_yml, expected_yml)
        self.assertEqual(result["updated"], False)
        self.assertDictEqual(result["yaml"], current_yml)

    def test_combine_column_lists_missing_columns(self):
        # Test when there are missing columns
        current_yml = {
            "name": "Model1",
            "columns": [
                {"name": "Column1", "type": "int"},
            ],
        }
        expected_yml = {
            "columns": [
                {"name": "Column1", "type": "int"},
                {"name": "Column2", "type": "str"},
            ]
        }
        result = datadict_yaml.combine_column_lists(current_yml, expected_yml)
        self.assertEqual(result["updated"], True)
        self.assertIn({"name": "Column2", "type": "str"}, result["yaml"]["columns"])

    def test_combine_column_lists_additional_columns(self):
        # Test when there are additional columns
        current_yml = {
            "name": "Model1",
            "columns": [
                {"name": "Column1", "type": "int"},
                {"name": "Column2", "type": "str"},
            ],
        }
        expected_yml = {
            "name": "Model1",
            "columns": [{"name": "Column1", "type": "int"}],
        }
        result = datadict_yaml.combine_column_lists(current_yml, expected_yml)
        self.assertEqual(result["updated"], True)
        self.assertDictEqual(result["yaml"], expected_yml)

    def test_combine_column_lists_empty_current_yml(self):
        # Test when current_yml has no columns initially
        current_yml = {"name": "Model1"}
        expected_yml = {
            "columns": [
                {"name": "Column1", "type": "int"},
            ]
        }
        result = datadict_yaml.combine_column_lists(current_yml, expected_yml)
        self.assertEqual(result["updated"], True)
        self.assertIn({"name": "Column1", "type": "int"}, result["yaml"]["columns"])


if __name__ == "__main__":
    unittest.main()



import logging
import subprocess

import ruamel.yaml


def parse_bash_outputs(input_string) -> str:
    """
    This function parses the bash output represented by `input_string`, extracts and returns the portion of the
    output starting from the occurrence of the substring 'version: 2' to the end.

    If 'version: 2' is not found, it returns an empty string. In case of any exceptions during the execution,
    it logs an error message with details of the exception.

    Args:
        input_string (str): The bash output as a string.

    Returns:
        str: The substring of `input_string` starting from 'version: 2' to the end. If 'version: 2' is not found,
             it returns an empty string.
    """
    try:
        version_two_index = input_string.find("version: 2")
        if version_two_index != -1:
            return input_string[version_two_index:]
        else:
            return ""
    except Exception as e:
        logging.error("There was an issue parsing the codegen outputs: " + e)


def validate_dbt() -> bool:
    """
    Validates the dbt project to ensure its integrity and required dependencies.

    This function performs the following checks to validate the dbt project:
    1. Runs `dbt debug` to check if the project passes all the debug checks. If there are any issues, it logs the
       encountered errors and returns False.
    2. Checks if `dbt-labs/codegen` is installed as a dependency using `dbt deps` command. If the required package is
       not found, it logs an error message and returns False.
    3. If both the above checks pass successfully, it logs a success message confirming the successful validation
       of the dbt project and returns True.

    Returns:
        bool: True if the dbt project is successfully validated; False otherwise.
    """
    logging.info("Validating dbt project...")
    try:
        # Check debug passes
        bash_command = ["dbt", "debug"]
        result = subprocess.run(bash_command, capture_output=True).stdout.decode(
            "UTF-8"
        )
        if "All checks passed!" not in result:
            logging.error(
                "Issues encountered when running `dbt debug`. Validate `dbt debug` passes before retrying."
            )
            return False

        # Check codegen installed
        bash_command = ["dbt", "deps"]
        result = subprocess.run(bash_command, capture_output=True).stdout.decode(
            "UTF-8"
        )
        if "dbt-labs/codegen" not in result:
            logging.error("dbt-labs/codegen is required to perform this operation")
            return False

        # Otherwise confirm valid
        logging.info("dbt project successfully validated")
        return True

    except Exception as e:
        logging.error("Issues encountered when attempting to validate dbt: " + e)
        return False


def get_model_yaml(model_names) -> str:
    """
    Generates the base model YAML for the specified model names.

    This function generates the base model YAML for the provided model names using dbt codegen's `generate_model_yaml`
    operation. The function performs the following steps:
    1. Logs an information message indicating the start of generating the base model YAML for the given model names.
    2. Constructs the arguments for the `dbt run-operation generate_model_yaml` command with the specified model names.
    3. Executes the command using `subprocess.run()` and captures the command's standard output as a string.
    4. Checks if there are any compilation errors in the output. If found, it logs an error message with the
       encountered issues and does not proceed further.
    5. If the output is error-free, it loads the output string into a YAML object using the `ruamel.yaml` library and
       returns the YAML content as a string.

    Parameters:
        model_names (list): A list of model names for which the base model YAML needs to be generated.

    Returns:
        str: The generated base model YAML as a string.

    Note:
        To use this function, the dbt CLI must be installed and accessible in the environment where this function is run.
    """
    try:
        logging.info(f'Generating base model for models: {", ".join(model_names)}')
        args = {"model_names": model_names}
        bash_command = [
            "dbt",
            "run-operation",
            "generate_model_yaml",
            "--args",
            str(args),
        ]
        result = subprocess.run(bash_command, capture_output=True).stdout.decode(
            "UTF-8"
        )
        if "Compilation Error" in result:
            logging.error(
                "Issues encountered when generating the model yaml: " + result
            )
        else:
            yaml = ruamel.yaml.YAML()
            return yaml.load(parse_bash_outputs(result))
    except Exception as e:
        logging.error(f"Issues encountered when generating the model yaml: {e}")



import logging
import os


def add_spaces_between_cols(file):
    """
    Add spaces between columns in a YAML file.

    This function reads a YAML file and inserts newlines between the columns that start with '- name:'.
    It then writes the modified content back to the same file.

    Parameters:
        file (str): The path to the YAML file to be processed.

    Returns:
        None
    """
    with open(file, "rt") as f:
        yaml = f.read()
    replaced = yaml.replace("dictionary:\n\n", "dictionary:\n").replace(
        "  - name:", "\n  - name:"
    )
    with open(file, "w") as f:
        f.write(replaced)


def open_model_yml_file(yaml_obj, file_path) -> dict:
    """
    Open and load a model YAML file for processing.

    This private method is used to open and load a YAML file from the provided file path. The function
    attempts to read and parse the file using the YAML parser. If the YAML data represents a valid model,
    it returns a dictionary with the status as "valid" and the loaded YAML data. Otherwise, it logs a
    message indicating that the file was skipped and returns a dictionary with the status as "invalid".

    Parameters:
        file_path (str): The path to the YAML file to be opened and loaded.

    Returns:
        dict: A dictionary with the keys "status" and "yaml". The "status" key will be either "valid" or "invalid".
            The "yaml" key will contain the loaded YAML data if valid, otherwise, it will contain None.
    """
    with open(file_path, "r+") as file:
        yaml = yaml_obj.load(file)
        if check_valid_model_file(yaml):
            return {"status": "valid", "yaml": yaml}
        else:
            return {"status": "invalid", "yaml": None}


def check_valid_model_file(model_yaml) -> bool:
    """
    Check if the parsed YAML data represents a valid model.

    This private method is used to check whether the parsed YAML data contains the required 'models'
    key, indicating that it represents a valid model file.

    Parameters:
        yaml (dict): The parsed YAML data.

    Returns:
        bool: True if the YAML data contains the required 'models' key, False otherwise.
    """
    try:
        _ = model_yaml["models"]
        return True
    except KeyError:
        return False


def output_model_file(yaml_obj, file_path, model_yaml, sort) -> None:
    """
    Output the updated model YAML data to a file.

    This private method is used to write the updated model YAML data to a file specified by the 'file_path'.
    The function takes the 'model_yaml' data as input and writes it to the file using the YAML serializer.

    Parameters:
        file_path (str): The path to the file where the updated model YAML should be written.
        model_yaml (dict): The updated model YAML data to be written to the file.

    Returns:
        None
    """
    if sort:
        output_yaml = sort_model_file(model_yaml)
        logging.info(f"File '{file_path}' has been sorted")
    else:
        output_yaml = model_yaml
    with open(file_path, "w") as file:
        yaml_obj.dump(output_yaml, file)
        logging.info(f"Updated model file '{file_path}'")


def list_directory_files(directory, extensions) -> dict:
    """
    Lists all files with the provided extensions in the specified directory and its subdirectories.

    This function recursively searches the provided 'directory' for files found with the supplied extensions.
    It returns a list containing the absolute file paths of all the found files.

    Parameters:
        directory (str): The path to the directory to search for files.
        extensions (list): A list of extensions to check for.

    Returns:
        list: A list containing absolute file paths of YAML files in the specified directory and its subdirectories.
    """
    try:
        files_list = []
        if os.path.exists(directory) and os.path.isdir(directory):
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if file.endswith(tuple(extensions)):
                        files_list.append(os.path.join(root, file))
            logging.info(
                f"Found {len(files_list)} files in the directory '{directory}' with extensions: {', '.join(extensions)}"
            )
        else:
            logging.error(f"Directory '{directory}' doesn't existing.")
        return files_list

    except Exception as e:
        logging.error(
            f"Issues encountered when trying to search directory for yaml files: {e}"
        )


def sort_model_file(file_yaml) -> dict:
    """
    Sorts the dictionary of models by the column names within each model alphabetically.

    This function takes a dictionary containing model information, where each model has a list of columns.
    The function sorts the columns within each model alphabetically based on their 'name' key.
    Additionally, the function sorts the models themselves alphabetically based on their 'name' key.

    Parameters:
        file_yaml (dict): The dictionary containing model information.

    Returns:
        dict: The sorted dictionary with models and their columns sorted alphabetically by their names.
    """

    # Sort the columns within each model alphabetically
    for model in file_yaml["models"]:
        if "columns" in model:
            model["columns"] = sorted(model["columns"], key=lambda col: col["name"])

    # Sort the models by name
    file_yaml["models"] = sorted(file_yaml["models"], key=lambda model: model["name"])

    return file_yaml



import click

import datadict


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "-d",
    "--dictionary",
    type=str,
    help="Location of the dictionary file",
    default="datadictionary.yml",
)
@click.option(
    "-D",
    "--directory",
    type=str,
    help="Directory to apply dictionary",
    default="models/",
)
def apply(dictionary, directory):
    """
    This command reviews all existing model files in the given directory for existing columns and collates them into a
    dictionary file. Additionally, this command will review the dictionary file and apply updates back to the columns in
    the model files where possible.
    """
    dictionary = datadict.datadict(dictionary, detailed_logs=True)
    dictionary.apply_data_dictionary_to_path(directory)
    dictionary.collate_output_dictionary()


@cli.command()
@click.option(
    "-D",
    "--directory",
    type=str,
    help="Directory to apply dictionary",
    default="models/",
)
@click.option(
    "-f",
    "--file",
    type=str,
    help="Name to give the generated YAML file",
    default="models.yml",
)
@click.option(
    "--unique-model-yaml",
    "unique_model_yaml",
    help="Creates one YAML for each model with the same name as the model",
    default=False,
)
@click.option(
    "--sort/--no-sort",
    help="Triggers the generated YAML files to be sorted alphabetically",
    default=True,
)

def generate(directory, file, unique_model_yaml, sort):
    """
    This command generates model YAML files in a specified directory. Existing model YAML files are evaluated,
    and the model metadata is combined and written back to the existing files. For models missing from existing files,
    a new file is created in the directory with the given name and the metadata for the missing models is written to it.
    """
    datadict.generate_model_yamls(directory, file, unique_model_yaml, sort)


import logging
import os

import ruamel.yaml

from datadict import datadict_helpers


class datadict:
    def __init__(self, dictionary_file_path, detailed_logs=True) -> None:
        """
        Initialize the object with the given dictionary file path and detailed logging settings.

        This constructor method is used to initialize an instance of the class. It sets the attributes
        'detailed_logs', 'dictionary_path', 'dictionary_yml', 'dictionary_items', 'existing_fields',
        and 'missing_fields' based on the provided inputs. The method also initializes logging and YAML
        configurations and loads the dictionary from the specified file.

        Parameters:
            dictionary_file_path (str): The file path to the YAML dictionary file.
            detailed_logs (bool, optional): Determines whether detailed log messages with 'info' level
                                            should be logged. Defaults to True.

        Returns:
            None
        """
        self.detailed_logs = detailed_logs
        self._init_logging()
        self._init_yaml()
        self.dictionary_path = dictionary_file_path
        self.dictionary_yml = self._format_dictionary(self._try_load_dictionary())
        self.dictionary_items = self._parse_aliases(self.dictionary_yml)
        self.existing_fields = []
        self.missing_fields = []

    def _init_yaml(self) -> None:
        """
        Initialize the YAML object and apply YAML configuration.

        This private method is used to initialize the YAML serializer object from the 'ruamel.yaml' library
        and apply specific configuration settings to it using the '_apply_yaml_config()' method.

        Parameters:
            None

        Returns:
            None
        """
        self.yaml = ruamel.yaml.YAML()
        self._apply_yaml_config()

    def _apply_yaml_config(self) -> None:
        """
        Apply YAML configuration settings.

        This private method is used to apply specific configuration settings to the YAML serializer in the
        object. It sets 'preserve_quotes' to True, which preserves quotes around strings in the output YAML.
        Additionally, it configures the indentation for mappings and sequences and sets the 'width' parameter
        for line wrapping in the output YAML.

        Parameters:
            None

        Returns:
            None
        """
        self.yaml.preserve_quotes = True
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.yaml.width = 200

    def _init_logging(self):
        """
        Initialize logging configuration for the object.

        This private method is used to set up the logging configuration for the object. It configures the
        logging level to INFO and specifies the format of the log messages to display the log level and the
        log message text.

        Parameters:
            None

        Returns:
            None
        """
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

    def _log(self, message, level="info") -> None:
        """
        Logs a message with the specified log level.

        This private method is used for logging messages with different log levels, such as 'info', 'warning',
        and 'error'. If 'detailed_logs' is set to True, messages with 'info' log level will also be logged.

        Parameters:
            message (str): The log message to be recorded.
            level (str, optional): The log level to use. Accepted values are 'info' (default), 'warning', and 'error'.

        Returns:
            None
        """
        if level == "warning":
            logging.warning(message)
        elif level == "error":
            logging.error(message)
        elif self.detailed_logs:
            logging.info(message)

    def _try_load_dictionary(self) -> dict:
        """
        Tries to load a dictionary from a given path, and if it doesn't exist, creates a new dictionary.

        This private method is used to load a dictionary from a specified path. If the file exists at the
        given path, it loads the dictionary using the `_load_dictionary()` method. If the file doesn't
        exist, it creates a new dictionary using the `_create_dictionary()` method.

        Returns:
            dict: A dictionary loaded from the specified path if the file exists, otherwise, a newly created dictionary.
        """
        if os.path.isfile(self.dictionary_path) and os.path.exists(
            self.dictionary_path
        ):
            self._log(f"Dictionary {self.dictionary_path} found successfully.")
            return self._load_dictionary()
        else:
            self._log(
                f"Dictionary {self.dictionary_path} not found. Creating dictionary..."
            )
            return self._create_dictinary()

    def _load_dictionary(self) -> dict:
        """
        Loads a dictionary from a specified path.

        This private method is used to read and load a dictionary from a given YAML file path. It attempts
        to open the file, read its content, and parse it as a YAML formatted dictionary.

        Returns:
            dict: The dictionary loaded from the specified YAML file.

        Raises:
            FileNotFoundError: If the file at the specified path does not exist.
            Exception: If an error occurs while loading the dictionary from the YAML file.
        """
        try:
            with open(self.dictionary_path, "r") as file:
                self._log(
                    f"Dictionary file '{self.dictionary_path}' loaded successfully."
                )
                return self.yaml.load(file)
        except FileNotFoundError:
            self._log(
                f"Dictionary file '{self.dictionary_path}' not found.", level="error"
            )
            raise
        except Exception:
            self._log(
                f"Error loading dictionary file '{self.dictionary_path}'", level="error"
            )
            raise

    def _create_dictinary(self) -> dict:
        """
        Creates a new dictionary and saves it to a specified path.

        This private method is used to create a new dictionary, serialize it as a YAML formatted data,
        and save it to a given file path. The created dictionary will have an initial structure of
        'base_yaml = {'dictionary': []}'.

        Returns:
            dict: The newly created dictionary.

        Raises:
            IOError: If an error occurs while creating or writing to the file.
            SystemExit: If a critical error occurs during the creation process.
        """
        try:
            with open(self.dictionary_path, "w") as file:
                base_yaml = {"dictionary": []}
                self.yaml.dump(base_yaml, file)
                self._log(
                    f"The file '{self.dictionary_path}' was successfully created."
                )
            with open(self.dictionary_path, "r") as file:
                self._log(
                    f"Dictionary file '{self.dictionary_path}' loaded successfully."
                )
                return self.yaml.load(file)
        except IOError:
            self._log(
                f"An error occurred while creating the file '{self.dictionary_path}'. Check the directory exists.",
                level="error",
            )
            raise SystemExit

    def _format_dictionary(self, dictionary_yml) -> dict:
        """
        Format the dictionary data to ensure consistent structure.

        This private method is used to format the YAML dictionary data to ensure that each field in the
        'dictionary' key contains 'description' and 'aliases' keys. If any field is missing the 'description'
        or 'aliases' keys, they will be added with appropriate default values. If the 'dictionary' key does
        not exist in the YAML data, it will be created with an empty list as the value.

        Parameters:
            dictionary_yml (dict): The YAML dictionary data to be formatted.

        Returns:
            dictionary_yml (dict): The formatted dictionary
        """
        try:
            if "dictionary" in dictionary_yml:
                if dictionary_yml["dictionary"] is not None:
                    for field_num, field in enumerate(dictionary_yml["dictionary"]):
                        if "description" not in field:
                            dictionary_yml["dictionary"][field_num]["description"] = ""
                        if "aliases" not in field:
                            dictionary_yml["dictionary"][field_num]["aliases"] = []
            else:
                dictionary_yml["dictionary"] = []
            return dictionary_yml
        except TypeError:
            self._log("There was an error when trying to format the dictionary")

    def _parse_aliases(self, dictionary) -> list:
        """
        Parse dictionary data to extract field names and their aliases.

        This private method is used to parse the YAML dictionary data and extract field names along with their
        associated aliases. The function searches for the 'dictionary' key in the provided 'dictionary' parameter,
        and if it exists, it iterates through each field to gather the field name and its aliases, if available.

        Parameters:
            dictionary (dict): The YAML dictionary data to be parsed.

        Returns:
            list: A list containing the field names and their aliases (if available).
        """
        try:
            values = []
            if dictionary["dictionary"] is None:
                return values
            for dict_column in dictionary["dictionary"]:
                values.append(dict_column["name"])
                try:
                    for alias in dict_column["aliases"]:
                        values.append(alias)
                except:
                    pass
            return values
        except TypeError:
            self._log("There was an error when trying to parse the dictionary")

    def _insert_dict_item(self, dictionary, key, value, index) -> dict:
        """
        Insert a new key-value pair into a dictionary at the specified index.

        This private method is used to insert a new key-value pair into the provided dictionary at the given index.
        The function first extracts the keys and values from the dictionary, then inserts the new key and value at
        the specified index. Finally, it creates a new dictionary with the modified key-value pairs and returns it.

        Parameters:
            dictionary (dict): The dictionary to which the new key-value pair should be inserted.
            key (hashable): The key to insert into the dictionary.
            value (any): The value associated with the new key to be inserted.
            index (int): The index at which the new key-value pair should be inserted.

        Returns:
            dict: A new dictionary with the inserted key-value pair at the specified index.
        """
        keys = list(dictionary.keys())
        values = list(dictionary.values())
        keys.insert(index, key)
        values.insert(index, value)
        modified_dict = dict(zip(keys, values))
        return modified_dict

    def _update_existing_field(self, model_column, model, file_path) -> None:
        """
        Update the list of existing fields with model column details.

        This private method is used to update the list of existing fields by appending information about a model
        column found in a model YAML file. The function takes the 'model_column', 'model', and 'file_path' as inputs.
        If the 'model_column' contains a 'description' key, it appends a dictionary with the column name, description,
        model name, and file path to the 'existing_fields' list. If the 'description' key is not present, it appends a
        dictionary without the 'description' key.

        Parameters:
            model_column (dict): The model column dictionary from the model YAML.
            model (dict): The model dictionary representing the current model from the YAML file.
            file_path (str): The file path of the YAML file containing the model.

        Returns:
            None
        """
        if "description" in model_column:
            self.existing_fields.append(
                {
                    "name": model_column["name"],
                    "description": model_column["description"],
                    "model": model["name"],
                    "file": file_path,
                }
            )
        else:
            self.existing_fields.append(
                {
                    "name": model_column["name"],
                    "model": model["name"],
                    "file": file_path,
                }
            )

    def _iterate_dictionary_update(self, model_yaml, file_path) -> dict:
        """
        Iterate through the model YAML and update dictionary fields if needed.

        This private method iterates through the model YAML and updates dictionary fields if they are found
        in the 'dictionary_yml'. For each model in the 'model_yaml', it checks if the model column name matches
        any entry in the 'dictionary_yml' or its aliases. If a match is found and the model YAML contains a
        'description' for that field, it updates the description from the 'dictionary_yml'. If the 'description'
        is missing, it inserts the 'description' key with the appropriate value.

        Parameters:
            model_yaml (dict): The model YAML dictionary to be updated.
            file_path (str): The file path of the YAML file containing the model.

        Returns:
            dict: A dictionary with keys "updated" and "model_yaml". "updated" will be True if any updates were made,
                False otherwise. "model_yaml" will contain the updated model YAML data.
        """
        updated = False
        try:
            for model_number, model in enumerate(model_yaml["models"]):
                if "columns" in model:
                    for col_num, model_column in enumerate(model["columns"]):
                        if self.dictionary_yml["dictionary"] is not None:
                            for dict_num, dict_column in enumerate(
                                self.dictionary_yml["dictionary"]
                            ):
                                if (
                                    model_column["name"] == dict_column["name"]
                                    or model_column["name"] in dict_column["aliases"]
                                ):
                                    if (
                                        "description"
                                        in model_yaml["models"][model_number][
                                            "columns"
                                        ][col_num]
                                    ):
                                        if (
                                            model_yaml["models"][model_number][
                                                "columns"
                                            ][col_num]["description"]
                                            != dict_column["description"]
                                            and dict_column["description"] != ""
                                        ):
                                            model_yaml["models"][model_number][
                                                "columns"
                                            ][col_num]["description"] = dict_column[
                                                "description"
                                            ]
                                            self._log(
                                                f"Field '{model_column['name']}' in file '{file_path}' has been updated."
                                            )
                                            updated = True
                                    elif dict_column["description"] != "":
                                        model_yaml["models"][model_number]["columns"][
                                            col_num
                                        ] = self._insert_dict_item(
                                            model_yaml["models"][model_number][
                                                "columns"
                                            ][col_num],
                                            "description",
                                            dict_column["description"],
                                            1,
                                        )
                                        self._log(
                                            f"Field '{model_column['name']}' in file '{file_path}' has been updated."
                                        )
                                        updated = True
                                    if "models" in dict_column:
                                        if (
                                            model["name"]
                                            not in self.dictionary_yml["dictionary"][
                                                dict_num
                                            ]["models"]
                                        ):
                                            self.dictionary_yml["dictionary"][dict_num][
                                                "models"
                                            ].append(model["name"])
                                    else:
                                        self.dictionary_yml["dictionary"][dict_num][
                                            "models"
                                        ] = [model["name"]]
                        self._update_existing_field(model_column, model, file_path)
                else:
                    self._log(
                        f"No columns found for model {model['name']} in '{file_path}'",
                        level="warning",
                    )
            if updated:
                return {"updated": True, "model_yaml": model_yaml}
        except Exception as error:
            self._log(
                f"Error getting file updates for '{file_path}': {error}", level="error"
            )
        return {"updated": False}

    def _collate_metadata(self, existing_fields) -> list:
        """
        Collates metadata from existing field list.

        This function takes a list of dictionaries representing existing fields and organizes the metadata
        by grouping fields based on their names. For each unique field name, it collects unique models and
        non-empty descriptions associated with the field.

        Parameters:
            existing_fields (list of dict): A list of dictionaries, where each dictionary contains information
                                            about an existing field with keys 'name', 'model', and optionally 'description'.

        Returns:
            list: A list of dictionaries containing collated metadata for each field. Each dictionary contains
                keys 'name', 'description', 'versions', and 'models'.
        """
        metadata = {}
        result = []

        # extract metadata from existing field list
        for field in existing_fields:
            name = field["name"]
            model = field["model"]
            description = field.get("description", "")

            if name not in metadata:
                metadata[name] = {
                    "description_versions": [description],
                    "description": description,
                    "models": [model],
                }
            else:
                metadata[name]["description_versions"].append(description)
                metadata[name]["models"].append(model)

        # summarise metadata
        for name, info in metadata.items():
            versions = list(
                set(
                    [
                        version
                        for version in info["description_versions"]
                        if version != ""
                    ]
                )
            )
            versions.sort()
            models = list(set(info["models"]))
            models.sort()
            if len(versions) > 1:
                result.append(
                    {
                        "name": name,
                        "description": "",
                        "description_versions": versions,
                        "models": models,
                    }
                )
            else:
                result.append(
                    {"name": name, "description": info["description"], "models": models}
                )

        # return field list sorted by name
        return sorted(result, key=lambda d: d["name"])

    def _output_dictionary(self) -> None:
        """
        Output the updated dictionary YAML data to a file.

        This private method is used to write the updated dictionary YAML data to a file specified by 'dictionary_path'.
        The function takes the 'dictionary_yml' data from the class instance and writes it to the file using the YAML
        serializer.

        Parameters:
            None

        Returns:
            None
        """
        try:
            with open(self.dictionary_path, "w") as file:
                self.yaml.dump(self.dictionary_yml, file)
                datadict_helpers.add_spaces_between_cols(self.dictionary_path)
            self._log(f"Dictionary '{self.dictionary_path}' has been updated")
        except Exception as error:
            self._log(
                f"There was a problem updating dictionary at '{self.dictionary_path}'. {error}",
                level="error",
            )

    def apply_data_dictionary_to_file(self, file_path) -> None:
        """
        Apply the data dictionary updates to the specified model YAML file.

        This method applies the data dictionary updates to the specified 'file_path' representing a model YAML file.
        It checks if the file contains valid model data by using '_open_model_yml_file' function. If the file is
        valid, it iterates through the model YAML data and updates the descriptions of fields based on the entries
        in the 'dictionary_yml'. If any updates are made, it writes the updated YAML data back to the file using the
        '_output_model_file' function. If no updates are made, it logs a message stating that no updates were found.

        Parameters:
            file_path (str): The path to the model YAML file to which the data dictionary updates should be applied.

        Returns:
            None
        """
        self._log(f"Checking file '{file_path}'...")
        model_yaml = datadict_helpers.open_model_yml_file(self.yaml, file_path)
        if model_yaml["status"] == "valid":
            try:
                updates = self._iterate_dictionary_update(model_yaml["yaml"], file_path)
                if updates["updated"]:
                    datadict_helpers.output_model_file(
                        self.yaml, file_path, updates["model_yaml"], False
                    )
                    self._log(f"File {file_path} has been updated")
                else:
                    self._log(f"No updates found for file '{file_path}'")

            except FileNotFoundError:
                self._log(f"File '{file_path}' not found.", level="error")
            except Exception as e:
                self._log(
                    f"Error processing file '{file_path}'. Error: " + e, level="error"
                )
        else:
            self._log(f"File '{file_path}' contains no models and has been skipped.")

    def apply_data_dictionary_to_path(self, directory) -> None:
        """
        Apply the data dictionary updates to all model YAML files in the specified directory and its subdirectories.

        This method applies the data dictionary updates to all model YAML files present in the specified 'directory'
        and its subdirectories. It iterates through the directory using os.walk and processes each YAML file using the
        'apply_data_dictionary_to_file' function.

        Parameters:
            directory (str): The path to the directory where model YAML files are located.

        Returns:
            None
        """
        if os.path.exists(directory) and os.path.isdir(directory):
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if file.endswith(".yaml") or file.endswith(".yml"):
                        file_path = os.path.join(root, file)
                        self.apply_data_dictionary_to_file(file_path)
        else:
            self._log(
                f"Directory '{directory}' doesn't exist or can't be found",
                level="error",
            )

    def collate_output_dictionary(self):
        """
        Collate metadata and update the data dictionary before writing to the dictionary file.

        This method is responsible for collating metadata from the 'existing_fields' list and updating
        the data dictionary ('dictionary_yml') with this information. The updated dictionary is then
        written back to the dictionary file specified during class initialization.

        Parameters:
            None

        Returns:
            None

        Behavior:
            1. The function first calls the '_collate_metadata(self.existing_fields)' method to collate
            metadata from the existing fields. The metadata contains information about the fields,
            including their names, descriptions, associated versions, and models.

            2. The metadata obtained in the previous step is then assigned to the 'dictionary' key of
            the class instance's 'dictionary_yml'. This key represents the dictionary data loaded
            from the YAML file.

            3. The function proceeds to write the updated 'dictionary_yml' to the dictionary file using
            the '_output_dictionary()' method.
        """
        existing_field_descriptions = self._collate_metadata(self.existing_fields)
        self.dictionary_yml["dictionary"] = existing_field_descriptions
        self._output_dictionary()



import logging
import os

import ruamel.yaml

from datadict import datadict_dbt, datadict_helpers


def check_files_for_models(yaml_obj, files) -> dict:
    try:
        file_yamls = []
        model_list = []
        for file_path in files:
            file_contents = datadict_helpers.open_model_yml_file(yaml_obj, file_path)
            if file_contents["status"] == "valid":
                try:
                    for model in file_contents["yaml"]["models"]:
                        model_list.append({"name": model["name"], "file": file_path})
                    file_yamls.append(
                        {"file_path": file_path, "file_yaml": file_contents["yaml"]}
                    )
                except Exception as error:
                    logging.error(
                        f"There was an issue processing file '{file_path}'. Ensure it is formatted correctly and retry. Error: {error}"
                    )
        return {"model_list": model_list, "file_yamls": file_yamls}
    except Exception as error:
        logging.error(
            f"Issues encountered when iterating through files to collect models. Error: {error}"
        )


def combine_column_lists(current_yml, expected_yml) -> dict:
    updated = False
    combined_yaml = current_yml.copy()
    existing_columns = combined_yaml.setdefault("columns", [])
    # Iterate through the columns of expected_yml and add the missing ones to current_yml
    for column in expected_yml.get("columns", []):
        name = column.get("name")
        if name is not None:
            # Check if the name is already present in current_yml
            found = False
            for existing_column in combined_yaml.get("columns", []):
                if existing_column.get("name") == name:
                    found = True
                    break

            # If the name is not already present, add the column to current_yml
            if not found:
                combined_yaml.setdefault("columns", []).append(column)
                updated = True
                logging.warning(
                    f"Missing column '{column['name']}' to be added to model '{current_yml['name']}'"
                )
       
	# Iterate through the existing columns and remove any that are not in the expected_yml
    columns_to_remove = []
    for existing_column in existing_columns:
        name = existing_column.get("name")
        if name is not None:
            found = False
            for column in expected_yml.get("columns", []):
                if column.get("name") == name:
                    found = True
                    break
            if not found:
                columns_to_remove.append(existing_column)

    if columns_to_remove:
        for column in columns_to_remove:
            existing_columns.remove(column)
            updated = True
            logging.warning(
                f"Column '{column['name']}' removed from model '{current_yml['name']}'"
            )
    
	# Add the data_type to any columns that are missing it & set empty description if missing
    data_types_dict = {column['name']: column['data_type'] for column in expected_yml.get("columns", [])}

    for column in combined_yaml['columns']:
        if 'data_type' not in column:
            column['data_type'] = data_types_dict[column['name']]
            updated = True
            logging.info(
			    f"Added data_type '{column['data_type']}' to '{column['name']}' in model '{current_yml['name']}'"
		    )

    sort_order = ['name', 'data_type', 'description', 'tests', 'data_tests', 'unit_tests', 'meta']
    
    for column in combined_yaml['columns']:
        column['description'] = column.get('description', '')

        for optional_columns in ['tests', 'data_tests', 'unit_tests', 'meta']:
            if optional_columns in column:
                column[optional_columns] = column[optional_columns]
            else:
                column.pop(optional_columns, None)

        column_sorted = {key: column[key] for key in sort_order if key in column}
        
        column.clear()
        column.update(column_sorted)
    
    return {"yaml": combined_yaml, "updated": updated}


def updated_existing_files(yaml_obj, existing_file_yamls, models_to_be_updated, sort):
    # loop through each existing file
    updated_count = 0
    for file in existing_file_yamls:
        file_yaml = file["file_yaml"]
        path = file["file_path"]

        try:
            for model_num, model in enumerate(file_yaml["models"]):
                    for model_to_be_updated in models_to_be_updated:
                        if model_to_be_updated["name"] == model["name"]:
                            logging.info(f"Model {model['name']} is being checked...")
                            combined_columns = combine_column_lists(
                                model, model_to_be_updated
                            )
                            file_yaml["models"][model_num] = combined_columns["yaml"]
                            updated = combined_columns["updated"]
                            if updated:
                                updated_count += 1
                            else:
                                logging.info(f"Model {model['name']} is correct")

            if updated_count > 0:
                    datadict_helpers.output_model_file(yaml_obj, path, file_yaml, sort)
        except Exception as error:
            logging.error(
                f"There was an issue processing file '{path}'. This is likely a badly formatted YAML file. Error: {error}"
            )


def add_missing_models(yaml_obj, path, models, sort):
    if os.path.isfile(path) and os.path.exists(path):
        yaml = datadict_helpers.open_model_yml_file(yaml_obj, path)
        if yaml["status"] == "valid":
            logging.info(f"File '{path}' has been found and is a valid models file")
            updated = yaml["yaml"]
            if len(models) > 0:
                updated["models"] = updated["models"] + models
                datadict_helpers.output_model_file(yaml_obj, path, updated, sort)
            else:
                logging.info(f"No updates to apply to '{path}'")
        else:
            logging.error(
                f"File '{path}' has been found and but isn't a valid model file"
            )
    else:
        logging.info(f"File '{path}' doesn't exist and will be created.")
        file_yaml = {"version": 2, "models": models}
        datadict_helpers.output_model_file(yaml_obj, path, file_yaml, sort)
        
def yaml_for_each_model(yaml_obj, model_file_list, existing_file_yamls, existing_model_list, models_to_be_added):
    """
    Check for each model if the current YAML file matches the expected YAML file. The expected YAML 
    file has the same name as the model file, but with a .yml extension.

    Parameters:
        yaml_obj (object): The YAML object.
        model_file_list (list): List of model file paths.
        existing_file_yamls (list): List of dictionaries with existing file YAMLs.
        existing_model_list (list): List of dictionaries with with existing models.
        models_to_be_added (list): List of dictionaries with models to be added.

    Returns:
        None
    """

    model_dicts = [{
        'name': os.path.splitext(os.path.basename(file))[0],
        'model_path': file,
        'expected_yml_path': file.replace('.sql', '.yml')
    } for file in model_file_list]

    existing_models = []

    for item in existing_file_yamls:
        if "models" in item["file_yaml"]:
            existing_models.extend(item["file_yaml"]["models"])
    
    combined_dicts = []

    for item1 in model_dicts:
        name = item1["name"]
        item2 = next((item for item in models_to_be_added if item["name"] == name), {})
        item3 = next((item for item in existing_model_list if item["name"] == name), {})
        item4 = next((item for item in existing_models if item["name"] == name), {})
        
        combined_dict = {**item1, **item2, **item3, **item4}
        combined_dicts.append(combined_dict)

    for dict in combined_dicts:
        dict.setdefault('file', None)
    
    files_to_remove = set()

    for model in combined_dicts:
        if model['file'] != model['expected_yml_path']:

            model_to_write = model.copy()
            model_to_write.pop('file', None)
            model_to_write.pop('model_path', None)
            model_to_write.pop('expected_yml_path', None)
            model_to_write = {
                'version': 2,
                'models': [model_to_write]
            } 

            datadict_helpers.output_model_file(yaml_obj, model['expected_yml_path'], model_to_write, sort=False)
            
            if model['file'] is not None:
                files_to_remove.add(model['file'])
                logging.info(f"Model '{model['name']}' is being split into its own yaml file.")

        else:
            logging.info(f"Model '{model['name']}' is correct")
            
    for file in files_to_remove:
        os.remove(file)


def generate_model_yamls(directory, name, unique_model_yaml, sort=True):
    """
    Generate model YAML files in a given directory.

    This function generates model YAML files in a specified directory. Existing model YAML files are evaluated,
    and the model metadata is combined and written back to the existing files. For models missing from existing files,
    a new file is created and the metadata for the missing models is written to it.

    Parameters:
        directory (str): The directory where the model YAML files are located, and where new files will be created.

        name (str): The base name of the new YAML file to be created for models missing from existing files.

        sort (bool, optional): Whether to sort the models alphabetically by their name. Default is False.

    Returns:
        None

    Raises:
        Exception: If any error occurs during the generation of the YAML files, an error message is logged and the
        exception is raised.

    Note:
        This function is part of a larger data dictionary process that requires a valid DBT (Data Build Tool)
        configuration and a set of SQL and YAML files containing DBT models and their metadata.
    """
    try:
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
        yaml_obj = ruamel.yaml.YAML()
        yaml_obj.preserve_quotes = True
        yaml_obj.indent(mapping=2, sequence=4, offset=2)
        yaml_obj.width = 200

        # 1. Validate dbt is configured and usable
        if not datadict_dbt.validate_dbt():
            return

        # 2. Evaluate the existing yaml files in the directory for model metadata
        yaml_file_list = datadict_helpers.list_directory_files(
            directory, [".yml", ".yaml"]
        )
        if len(yaml_file_list) == 0:
            SystemExit
        existing_files = check_files_for_models(yaml_obj, yaml_file_list)
        existing_model_list = existing_files["model_list"]
        existing_file_yamls = existing_files["file_yamls"]

        # 3. Get the full column list for every model in the directory
        model_file_list = datadict_helpers.list_directory_files(directory, [".sql"])
        model_names = [os.path.basename(file).split(".")[0] for file in model_file_list]
        model_column_list = datadict_dbt.get_model_yaml(model_names)

        # 4. Split out the models in existing files from models missing from existing files.
        models_to_be_updated = []
        models_to_be_added = []

        for model_num, model in enumerate(model_column_list["models"]):
            if any(
                existing_model["name"] == model["name"]
                for existing_model in existing_model_list
            ):
                models_to_be_updated.append(model_column_list["models"][model_num])
            else:
                models_to_be_added.append(model_column_list["models"][model_num])

        # 5. For models in existing files, combine the column lists and write back to the existing file 
        if len(models_to_be_updated) > 0:
            logging.info(f"There are {len(models_to_be_updated)} models to be checked")
            updated_existing_files(
                yaml_obj, existing_file_yamls, models_to_be_updated, sort
            )
        else:
            logging.info("There are no models requiring updating.")

        # 6. For models missing from existing files, create a new file with the given name and output the metadata
        if unique_model_yaml:
            yaml_for_each_model(yaml_obj, model_file_list, existing_file_yamls, existing_model_list, models_to_be_added)
        elif len(models_to_be_added) > 0:
            file_name = os.path.join(directory, name)
            logging.info(
                f"There are {len(models_to_be_added)} models to be added to file '{file_name}'"
            )
            add_missing_models(yaml_obj, file_name, models_to_be_added, sort)
        else:
            logging.info("There are no models to be added.")

    except Exception as error:
        logging.error(f"There was an error generating the YAML files. Error: {error}")
